#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-05-30 10:28:49
# @Description :  some useful functions of indexes or indicators that measure
#                   the degree of diversity.

__version__ = "0.1.1"

import math


def category_count(data: list):
    """学科数量丰富度variety，即category count（MacArthur 1965）

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        int: 学科数量，也可以理解为分类的数量
    """
    return len(data)


# def disparity(data: list):
#     """学科差异度，（Weizman(1992a)、Solow & Polasky(1994a)
#
#     Args:
#         data(list): 列表中存放不同（物种）分类的出现次数
#
#     Returns:
#         int: 返回差异性指标
#     """
#     pass


def shannon_evenness(data: list):
    """学科分布均匀度balance，即Shannon evenness(Pielou 1969)，又称为Shannon’s equitability

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        float: Shannon evenness的数值，衡量社区（当前区域/领域)的物种均匀程度

    """
    si = shannon_index(data)

    # Maximum diversity possible
    h_max = math.log(category_count(data))

    return float(si) / h_max if h_max != 0 else 0


def shannon_index(data: list):
    """考虑了variety/balance，Shamono & Weaver（1962）。

    Shannon指数是生态文献中比较流行的多样性指数，
    也叫Shannon Diversity Index、Shannon-Wiener指数、Shannon-Weaver指数和Shannon Entropy。

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        float: Shannon Diversity Index
    """

    def p(n, N):
        """
        计算相对丰富度

        Args:
            n(int): 每个物种（分类）的数目
            N(int): 所有物种（分类）的总数

        Returns:
            float: 相对丰度值
        """
        if N == 0:
            ratio = 0
        else:
            ratio = float(n) / N
        return ratio * math.log(ratio) if ratio != 0 else 0

    N = sum(data)
    return -sum(p(n, N) for n in data)


def simpson_index(data: list):
    """考虑了 variety/balance，Simpson（1949），有关Simpson Index的名称（叫法），
    在不同的地方有不同的形式，请参考链接 http://www.countrysideinfo.co.uk/simpsons.htm。

    Simpson Index 衡量个体被分类时的集中程度，描述从一个群落中连续两次抽样所得到的个体属于同一种的概率
    SI 的值域为[0, 1]，0意味着无限多样性，1意味着没有多样性，即值越大，多样性越小。
    由于该形式的定义通常成反比，因此，通常使用 Inverse Simpson Index 和 Gini-Simpson Index。

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        float: Simpson Index 的数值
    """
    N = sum(data)
    si = sum((float(n) / N if N != 0 else 0) ** 2 for n in data)
    return si


def inverse_simpson_index(data: list):
    """Inverse Simpson Index，即逆辛普森指数，也叫做Simpson's Reciprocal Index。

    ISI指数的值，最小值为1，表示多样性最小，只有一个物种的群落。
    其最大值为样本中的物种数目。如果样本中有5个物种，则最大值为5。

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        float: Inverse Simpson Index 的数值
    """
    si = simpson_index(data)
    return float(1) / si if si != 0 else 0


def gini_simpson_index(data: list):
    """考虑了variety/balance，Simpson（1949），即Gini-Simpson Index，
    也称为Simpson Diversity Index，衡量物种多样性的指数。
    指从一个群落中连续两次抽样所得到的个体不属于同一种的概率，该值越大，说明该群落的多样性越高。

    GSI的值域为[0, 1]，值越大，表示多样性越高。

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        float: Gini-Simpson Index 的数值
    """
    return 1 - simpson_index(data)


def brillouin_diversity_index(data: list):
    """计算布里渊多样性指数。如，计算作者A的BI指数，传入[30,20,10,10]，
    表示该作者在4个学科领域上的reference次数

    Args:
        data(list): 列表中存放不同（物种）分类的出现次数

    Returns:
        float: brillouin index的值
    """
    sum_ = sum(data)
    a = math.factorial(sum_)
    A = math.log10(a)

    bs = [math.log10(math.factorial(n)) for n in data]
    B = sum(bs)

    H = (A - B) / sum_ if sum_ != 0 else 0

    return H


if __name__ == "__main__":

    data = [10, 20, 30]
    data = [10, 20, 30, 0]

    print(category_count(data))
    print(shannon_evenness(data))
    print(shannon_index(data))
    print(simpson_index(data))
    print(inverse_simpson_index(data))
    print(gini_simpson_index(data))
    print(brillouin_diversity_index(data))
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-10-29 13:38:36
# @Description :  Powered by GitOPEN

__version__ = "0.2.7"

import json
import os
import time
from pathlib import Path, PosixPath
from types import FunctionType

from gitopenlib.utils import basics as gb


def get_paths_from_dir(dirs: str or list, types: str or list, recusive: bool = False):
    """
    从指定目录下获取所有指定扩展名文件的路径，不递归子文件夹

    Args:
        dirs (list): 文件夹路径（绝对路径），单个用str表示，多个用list
        types (types): 指定文件的扩展名，单个用str表示，多个用list
        recusive (bool): 是否递归子文件，默认为False

    Returns:
        list: 文件路径字符串列表
    """

    if isinstance(dirs, str):
        dirs = [dirs]
    if isinstance(types, str):
        types = [types]

    rule = "**/*." if recusive else "*."
    result = []
    for d in dirs:
        path = Path(d).resolve()
        for t in types:
            res = path.glob(rule + t)
            result.extend(list(res))

    return [str(item) for item in result]


def file_writer(
    lines: str or list,
    dir_path: str or Path,
    file_name: str = "file_writer.txt",
    mode: str = "a+",
    separator: str = "\n",
    encoding: str = "utf-8",
    backup: bool = True,
):
    """向文件中写内容

    Args:
        lines (str or list): 可以是单个字符串或者字符串的列表
        dir_path (str or Path): 文件的绝对路径
        file_name (str): 文件名称，有默认值
        mode (str): 写文件的的模式，默认为 a+
        separator (str): 每一行末尾的分隔符，有默认值
        encoding (str): 文件的编码格式，默认为utf-8
        backup (bool): 如果文件已经存在，是否对原文件进行备份，默认为True；
            如果为False，当文件已经存在的情况下，会追加到存在的文件中。

    Returns:
        None: 无返回值
    """
    dir_path: Path = Path(dir_path) if isinstance(dir_path, str) else dir_path

    if not dir_path.exists():
        print(
            "## Warning: The dir path you submit is not exist, "
            + "it will be created automatically."
        )
        dir_path.mkdir(parents=True, exist_ok=True)

    file_path = dir_path / file_name

    if backup:
        if file_path.exists():
            file_path.rename(
                file_path.with_suffix(
                    f'.{str(time.strftime("%Y%m%d_%H%M%S",time.localtime()))}{file_path.suffix}'
                )
            )

    if isinstance(lines, str):
        lines = [lines]

    with open(file_path, mode=mode, encoding=encoding) as file:
        for line in lines:
            if separator != "":
                file.write(f"{line}{separator}")
            else:
                file.write(line)

    print("## file_writer done...")
    pass


def read_txt_by_page(
    file_path: str,
    parse_func: FunctionType,
    page_size: int = 100,
    encoding: str = "utf-8",
):
    """
    从文本文件中分页读取内容

    Args:
        file_path (str): 文件路径
        parse_func (FunctionType): 每一页数据处理函数，必须定义
        page_size (int): 每一页数据量
        encoding (str): 文本文件的编码格式

    Returns:
        None: 在parse_func已经处理好数据，不用返回值
    """
    with open(file=file_path, encoding=encoding, mode="r") as file:
        data = list()
        curr_page_id = 0
        start_time = time.time()
        total_time = 0.0
        print("## read_txt_by_page is starting parse the data...")
        for line in file:
            data.append(line.strip())
            if len(data) == page_size:
                parse_func(data)
                data.clear()
                end_time = time.time()
                cost_time = float(end_time - start_time)
                total_time += cost_time
                start_time = end_time
                print(f"## -{curr_page_id}- page parsed done...{[cost_time]}s")
                curr_page_id += 1

            pass
        if len(data) > 0:
            parse_func(data)
            end_time = time.time()
            cost_time = float(end_time - start_time)
            total_time += cost_time
            start_time = end_time
            print(f"## -{curr_page_id}- page parsed done...{[cost_time]}s")
            curr_page_id += 1
    print(f"## All done. Total pages: {curr_page_id}. Elapsed time: {total_time}s")


def read_content(file_path: str or PosixPath, encoding: str = "utf-8"):
    """
    从文本文件中读取内容，将内容转换为list，list的元素为每行的字符串

    Args:
        file_path (str or PosixPath): 文件路径
        encoding (str): 文件的编码方式

    Returns:
        list: 每行字符串组成的列表
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    return gb.remove_0_str(
        [line.strip() for line in file_path.read_text(encoding=encoding).split("\n")]
    )


def read_jsons(file_path: str or PosixPath, encoding: str = "utf-8"):
    """
    从存放json的文本文件中读取内容，并转化为dict组成的list

    Args:
        file_path (str): 文件路径
        encoding (str): 文件的编码方式

    Returns:
        list: dict组成的list
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    result = gb.remove_0_str(
        [line.strip() for line in file_path.read_text(encoding=encoding).split("\n")]
    )

    return [json.loads(item) for item in result]


if __name__ == "__main__":

    #  def parse(data):
    #      time.sleep(0.5)
    #
    #  file_path = "/Users/sunjiajia/Works/Projects/PycharmProjects/interdisciplinary2/output/results/test_cits_10000.txt"
    #  read_txt_by_page(file_path=file_path, parse_func=parse, page_size=1000)

    file_writer(
        lines=["1", "adb", "sdf"],
        dir_path="~/Downloads/test9919/tere",
        file_name="xxx.txt",
        backup=False,
    )
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-05-30 22:51:45
# @Description : 包含基本的文件读写，指定扩展名文件查找等基本工具


__version__ = "0.2.0"


import json
import math


def ele2dict(data: list):
    """把元素为json字符串的list数据，转为，元素为dict类型的list数据

    Args:
        data (list): list类型的数据，元素为json字符串类型

    Returns:
        list : 转换后的list数据
    """
    return [json2dict(item) for item in data]


def ele2json(data: list):
    """把元素为dict类型的list数据，转为，元素为json字符串的list数据

    Args:
        data (list): list类型的数据，元素为dict类型

    Returns:
        list : 转换后的list数据
    """
    return [dict2json(item) for item in data]


def dict2json(data: dict):
    """将dict转为json字符串

    Args:
        data (dict): dict类型的数据

    Returns:
        str : 返回json字符串
    """
    return json.dumps(data, ensure_ascii=False)


def json2dict(astr: str):
    """将json字符串转为dict类型的数据对象

    Args:
        astr (str): json字符串转为dict类型的数据对象

    Returns:
        str : 返回dict类型数据对象
    """
    return json.loads(astr)


def dict_sorted(data: dict, flag: int = 0, ascending: bool = True):
    """
    对dict排序

    Args:
        data (dict): 目标数据dict类型
        flag (int): 默认为0，表示按照字典的key进行排序；1表示按照value进行排序
        ascending (bool): 默认为True，表示按照升序排序；False表示降序排序

    Returns:
        dict: 排序后的数据
    """
    return dict(sorted(data.items(), key=lambda x: x[flag], reverse=not ascending))


def list_deduplicate(data: list):
    """列表项为dict类型的去重

    Args:
        data (list): 目标数据，list类型

    Returns:
        list: 去重后的数据
    """
    return [dict(t) for t in set([tuple(d.items()) for d in data])]


def sort_list(
    data: list,
    ascending: bool = True,
    flag: int = 0,
    position: int = 0,
    key: str = "",
):
    """
    对 list 进行排序

    Args:
        data (list): list 类型数据，item 为 dict 或者 basic type
        ascending (bool): 默认为 True，表示升序；False 表示降序
        flag (int): 0 表示元素为基本类型，1 表示元素为 dict，2 表示元素为 tuple
        position (int): 如果元素为 basic type，position 保持默认即可；
            如果元素为 tuple，position 的数值表示以哪个 index 位置的值排序
        key (str): 如果元素为 dict，key 表示按照哪个 key 的 value 进行排序

    Returns:
        list: 排序后的数据
    """
    if flag == 0:
        # basic type
        data.sort(reverse=not ascending)
    elif flag == 1:
        # dict type
        if key == "":
            key, _ = data[0].items()[0]
        data.sort(key=lambda x: x.get(key, 0), reverse=not ascending)
    elif flag == 2:
        # tuple type
        if position < 0 or position >= len(data[0]):
            position = 0
        data.sort(key=lambda x: x[position], reverse=not ascending)

    return data


def strips(string: str):
    """
    去除字符串两端的空格符和换行符，并且去除中间的换行符
    """
    return string.strip().replace("\n", "").replace("\r", "").replace("\r\n", "")


def remove_0_str(data: list):
    """
    去除list列表中为长度为0的字符串，用于字符串split后，列表中出现长度为0字符串的去除
    """
    return [item for item in data if len(str(item)) != 0]


def chunks(arr, m):
    """分割列表，但是子list元素个数尽可能平均

    Args:
        arr (list): 待分割的list
        m (int): 分成几份

    Returns:
        list: 分割后的每个子list都是返回结果list的一个元素
    """
    n = int(math.ceil(len(arr) / float(m)))
    return [arr[i : i + n] for i in range(0, len(arr), n)]



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-05-30 22:52:49
# @Description :  基于Liu Huan进行简单的改写，改写为类。感谢Liu Huan（liuhuan@mail.las.ac.cn)


__version__ = "0.1.0"


import numpy as np
import pandas as pd


class CoMatrix:
    def __init__(self, target: list, normal=True):
        self.target = target
        self.normal = normal

    def authors_stat(self, target):
        au_dict = {}  # 单个关键词频次统计
        au_group = {}  # 两两关键词合作
        for authors in target:
            authors = authors.split(",")  # 按照逗号分开每个关键词
            authors_co = authors  # 合关键词同样构建一个样本
            for au in authors:
                # 统计单个关键词出现的频次
                if au not in au_dict:
                    au_dict[au] = 1
                else:
                    au_dict[au] += 1
                # 统计合作的频次
                authors_co = authors_co[1:]  # 去掉当前关键词
                for au_c in authors_co:
                    A, B = au, au_c  # 不能用本来的名字，否则会改变au自身
                    if A > B:
                        A, B = B, A  # 保持两个关键词名字顺序一致
                    co_au = A + "," + B  # 将两个关键词合并起来，依然以逗号隔开
                    if co_au not in au_group:
                        au_group[co_au] = 1
                    else:
                        au_group[co_au] += 1
        return au_group, au_dict

    def generate_matrix(self, au_group, au_dict):
        # 取出所有单个关键词
        au_list = list(au_dict.keys())
        # 新建一个空矩阵
        matrix = pd.DataFrame(np.identity(len(au_list)), columns=au_list, index=au_list)
        for key, value in au_group.items():
            A = key.split(",")[0]
            B = key.split(",")[1]
            Fi = au_dict[A]
            Fj = au_dict[B]
            if self.normal:
                Eij = value * value / (Fi * Fj)
            else:
                Eij = value
            # 按照关键词进行索引，更新矩阵
            matrix.loc[A, B] = Eij
            matrix.loc[B, A] = Eij
        return matrix

    def to_tsv(self, path_csv, path_tsv):
        csv_file = open(path_csv, "r")
        tsv_file = open(path_tsv, "w+")
        for index, line in enumerate(csv_file):
            if index == 0:
                line = "*" + line
            line = line.replace(",", "\t")
            tsv_file.write(line)
        tsv_file.close()
        csv_file.close()

    def run(self):
        au_group, au_dict = self.authors_stat(self.target)
        # print("au_group:", au_group)
        # print("au_dict:", au_dict)
        matrix = self.generate_matrix(au_group, au_dict)
        # print(matrix)
        # 相关系数
        # print(matrix.corr())
        if self.normal:
            file_name = "matrix_normal"
        else:
            file_name = "matrix"
        matrix.to_csv(f"{file_name}.csv", index_label="id")
        print(f"the {file_name}.csv is saved..")
        self.to_tsv(f"{file_name}.csv", f"{file_name}.tsv")
        print(f"the {file_name}.tsv is saved..")


#  if __name__ == "__main__":
#      kws = ["a,b,c", "123,123,456"]
#      cm = CoMatrix(target=kws, normal=True)
#      cm.run()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-11-05 10:23:57
# @Description :  一系列统计学相关的计算函数

from collections import Counter

import numpy as np

__version__ = "0.3.0"


def calculate_list_count_percent(data: list, decimals: None or int = None):
    """计算list中元素的次数和百分比

    Args:
        data (list): list类型数据。
        decimals (None or int): 保留的小数位数，默认为None，则不进行位数操作；
            如果>=0，则保留相应的小数位数。

    Returns:
        tuple: 元组第一个元素为次数统计结果，第二个元素为百分比统计结果。

    """
    counter = dict(Counter(data))

    total_count = sum(list(counter.values()))

    percenter = dict(
        [
            (
                key,
                value / total_count
                if decimals is None
                else round(value / total_count, decimals),
            )
            for key, value in counter.items()
        ]
    )
    return counter, percenter


def smooth_ma(data: list or np.array, window_size=3):
    """
    平滑后的数组的长度为：len(data) - window_size + 1

    Args:
        data (list or np.array): 原始数组。
        window_size (int): 窗口大小。

    Returns:
        np.array: 平滑后的数组
    """
    cumsum_vec = np.cumsum(np.insert(data, 0, 0))
    ma_vec = (cumsum_vec[window_size:] - cumsum_vec[:-window_size]) / window_size
    return ma_vec


def smooth_matlab(data: list or np.array, window_size: int = 3):
    """
    这个方法是matlab中平滑函数的python实现。它的窗口只能为奇数。
    平滑后的数组的长度为：len(data)

    Args:
        data (list or np.array): 原始数组。
        window_size (int): 窗口大小。

    Returns:
        np.array: 平滑后的数组

    """

    # a: NumPy 1-D array containing the data to be smoothed
    # WSZ: smoothing window size needs, which must be odd number,
    # as in the original MATLAB implementation
    out0 = np.convolve(data, np.ones(window_size, dtype=int), "valid") / window_size
    r = np.arange(1, window_size - 1, 2)
    start = np.cumsum(data[: window_size - 1])[::2] / r
    stop = (np.cumsum(data[:-window_size:-1])[::2] / r)[::-1]
    return np.concatenate((start, out0, stop))


def normalization(data: list or np.array, decimals: None or int = None):
    """
    对一系列数据进行归一化处理，对原始数据进行线性变换把数据映射到[0,1]之间。

    Args:

        data (list or np.array): 需要被归一化的数据。
        decimals (None or int): 归一化后，数值保留的精度（小数位数），默认为None，不开启

    Returns:
        np.array: 处理后的数据
    """
    _range = np.max(data) - np.min(data)
    ret = (data - np.min(data)) / _range
    return ret if decimals is None else np.around(ret, decimals=decimals)


def standardization(data: list or np.array, decimals: None or int = None):
    """
    对一系列数据进行标准化处理，常用的方法是z-score标准化，处理后数据均值为0，标准差为1。

    Args:

        data (list or np.array): 需要被标准化的数据。
        decimals (None or int): 归一化后，数值保留的精度（小数位数），默认为None，不开启

    Returns:
        np.array: 处理后的数据
    """
    mu = np.mean(data, axis=0)
    sigma = np.std(data, axis=0)
    ret = (data - mu) / sigma
    return ret if decimals is None else np.around(ret, decimals=decimals)


def zero_centered(data: list or np.array, decimals: None or int = None):
    """
    对一系列数据进行零均值化，即zero-centered。对数据的平移的一个过程，之后数据的中心点为(0,0)。

    Args:

        data (list or np.array): 需要被零均值化的数据。
        decimals (None or int): 数值保留的精度（小数位数），默认为None，不开启

    Returns:
        np.array: 处理后的数据
    """
    mu = np.mean(data, axis=0)
    ret = data - mu
    return ret if decimals is None else np.around(ret, decimals=decimals)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-06-14 11:12:51
# @Description :
# helpers意味“帮助者”，主要存放可以简化操作的数据库（pymongo、mysql）等操作。



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-06-01 12:05:26
# @Description :  10种排序的方法，初步测试结果（仅供参考），quick_sort < insert_sort <
# count_sort < shell_sort < merge_sort < bucket_sort < radix_sort < heap_sort <
# select_sort < bubble_sort


def bubble_sort(arr):
    """冒泡排序

    bubble_sort 耗时：284314 毫秒
    """
    n = len(arr)
    for i in range(n - 1):
        for j in range(0, n - 1 - i):
            if arr[j] > arr[j + 1]:
                arr[j], arr[j + 1] = arr[j + 1], arr[j]
    return arr


def quick_sort(listt, left, right):
    """快速排序

    当left=0, right=1000时，耗时 2 毫秒
    """
    if left >= right:
        return listt

    # 选择参考点，该调整范围的第1个值
    pivot = listt[left]
    low = left
    high = right

    while left < right:
        # 从右边开始查找大于参考点的值
        while left < right and listt[right] >= pivot:
            right -= 1
        # 这个位置的值先挪到左边
        listt[left] = listt[right]

        # 从左边开始查找小于参考点的值
        while left < right and listt[left] <= pivot:
            left += 1
        # 这个位置的值先挪到右边
        listt[right] = listt[left]

    # 写回改成的值
    listt[left] = pivot

    # 递归，并返回结果
    quick_sort(listt, low, left - 1)  # 递归左边部分
    quick_sort(listt, left + 1, high)  # 递归右边部分
    return listt


def insert_sort(alist):
    """插入排序

    insert_sort 耗时：16 毫秒
    """
    length = len(alist)
    for i in range(1, length):
        for j in range(i, 0, -1):
            if alist[j] < alist[j - 1]:
                alist[j], alist[j - 1] = alist[j - 1], alist[i]
            else:
                break
    return alist


def shell_sort(alist):
    """希尔排序

    shell_sort 耗时：150 毫秒
    """
    n = len(alist)
    gap = n // 2
    while gap >= 1:
        for i in range(gap, n):
            while (i - gap) >= 0:
                if alist[i] < alist[i - gap]:
                    alist[i], alist[i - gap] = alist[i - gap], alist[i]
                    i = i - gap
                else:
                    break
        gap = gap // 2
    return alist


def select_sort(alist):
    """选择排序

    select_sort 耗时：121566 毫秒
    """
    n = len(alist)
    for i in range(n):
        # 设置索引 i为最小值的索引
        min_idx = i
        # 通过比较，不断调整最小值的索引位置
        for j in range(i, n):
            if alist[j] < alist[min_idx]:
                min_idx = j
        # 交换第i个值与最小值
        alist[i], alist[min_idx] = alist[min_idx], alist[i]
    return alist


def heap_sort(heap):
    """堆排序，将根节点取出与最后一位对调，对前面len-1个节点继续进行对调过程

    heap_sort 耗时：520 毫秒
    """

    def max_heap(heap, heapsize, root):
        """
        调整堆的结构，使其父节点的值大于子节点的值
        """
        left = 2 * root + 1
        right = left + 1
        large = root
        if left < heapsize and heap[large] < heap[left]:
            large = left
        if right < heapsize and heap[large] < heap[right]:
            large = right
        # 若large=right或large=left，则说明，出现比父节点大的子节点，这时对调，使子节点变为父节点
        if large != root:
            heap[large], heap[root] = heap[root], heap[large]
            max_heap(heap, heapsize, large)

    def build_max_heap(heap):
        """
        构造一个堆，对堆中数据重新排序
        """
        length = len(heap)
        # 从后往前调整结构
        for i in range((length - 2) // 2, -1, -1):
            max_heap(heap, length, i)

    build_max_heap(heap)
    for i in range(len(heap) - 1, -1, -1):
        heap[0], heap[i] = heap[i], heap[0]
        max_heap(heap, i, 0)
    return heap


def merge_sort(alist):
    """归并排序

    merge_sort 耗时：239 毫秒
    """

    def merge(left, right):
        i = 0
        j = 0
        result = []
        while i < len(left) and j < len(right):
            if left[i] < right[j]:
                result.append(left[i])
                i += 1
            else:
                result.append(right[j])
                j += 1
        result += left[i:]
        result += right[j:]
        return result

    if len(alist) < 2:
        return alist
    # 将列表分成更小的两个列表
    mid = int(len(alist) / 2)
    # 分别对左右两个列表进行处理，分别返回两个排序好的列表
    left = merge_sort(alist[:mid])
    right = merge_sort(alist[mid:])
    # 对排序好的两个列表合并，产生一个新的排序好的列表
    return merge(left, right)


def count_sort(alist):
    """技术排序

    count_sort 耗时：47 毫秒
    """
    # 找到最大最小值
    min_num = min(alist)
    max_num = max(alist)
    # 初始化计数列表
    count_list = [0] * (max_num - min_num + 1)
    # 对列表中的每一个元素计数
    for num in alist:
        count_list[num - min_num] += 1
    alist.clear()
    # 当某个元素的个数不为 0，将该元素填回alist列表
    for cur_num, count in enumerate(count_list):
        while count != 0:
            alist.append(cur_num + min_num)
            count -= 1
    return alist


def bucket_sort(alist):
    """桶排序

    bucket_sort 耗时：290 毫秒
    """
    min_num = min(alist)
    max_num = max(alist)
    # 设置桶的大小
    bucket_size = (max_num - min_num) / len(alist)
    # 创建桶数组
    bucket_list = [[] for i in range(len(alist) + 1)]
    # 向桶数组填数
    for num in alist:
        bucket_list[int((num - min_num) // bucket_size)].append(num)
    alist.clear()
    # 回填，这里桶内部排序直接调用了sorted
    for i in bucket_list:
        for j in sorted(i):
            alist.append(j)
    return alist


def radix_sort(alist):
    """基数排序

    radix_sort 耗时：439 毫秒
    """
    # 记录正在对哪一位进行排序，最低位为个位
    i = 0
    # 最大值的位数
    max_num = max(alist)
    j = len(str(max_num))
    while i < j:
        # 建立桶数组，数字为0-9，所以建10个桶
        bucket_list = [[] for i in range(10)]
        # 按位数的大小分别装进不同的桶里
        for num in alist:
            bucket_list[int(num / (10 ** i) % 10)].append(num)
        # 将原列表清空，将各个桶里的数据依次添加进原列表
        alist.clear()
        for l in bucket_list:
            for b in l:
                alist.append(b)
        # 再进行前一位的排序，依次循环，直到排序的位数大于最大值的位数
        i += 1
    return alist


# def test(arr, func):
# """
# 测试排序函数的时间效率
# """
# start_time = time.time()
# if func.__name__ == "quick_sort":
#   func(arr, 0, 1000)
# else:
#   func(arr)
# end_time = time.time()
# print(
#   "{} 耗时：{} 毫秒".format(func.__name__, int(round((end_time - start_time) * 1000)))
# )
# print("-" * 128)


#  if __name__ == "__main__":
#      # import nump as np
#      # import time
#
#      # arr = list(np.random.randint(1, 100000, (50000)))
#      # test(arr, bubble_sort)
#      # test(arr, bucket_sort)
#      # test(arr, count_sort)
#      # test(arr, heap_sort)
#      # test(arr, insert_sort)
#      # test(arr, merge_sort)
#      # test(arr, quick_sort)
#      # test(arr, radix_sort)
#      # test(arr, select_sort)
#      # test(arr, shell_sort)
#
#      pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-06-01 11:42:36
# @Description : 存放常用的工具，包含：基本工具（文件读写）、排序函数（10种排序）




#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-05-30 22:49:57
# @Description :  提供一系列的有关操作mongodb/pymongo的工具


__version__ = "0.1.2.13"


import asyncio
import time
from typing import Callable

import pymongo
from bson.objectid import ObjectId
from pymongo.client_session import ClientSession
from pymongo.collection import Collection

from gitopenlib.utils import basics as gb


class ManageDB:
    """
    A simple manager class for pymongo
    """

    def __init__(self, host="127.0.0.1", port=27017):
        self.host = host
        self.port = port
        self._client = pymongo.MongoClient(host=self.host, port=self.port)

    def client(self):
        """
        get the mongodb client.
        """
        return self._client

    def coll(self, db_name, collection_name):
        """
        get the collection from specific db.
        """
        return self._client[db_name][collection_name]


def find_by_page(coll, page_size, parse_func):
    """
    find the data by page and process it through the parse function.

    Args:

        coll(Collection):   collection object.
        page_size(int):  the page size.
        parse_func: A handler function, with a parameter of type list, implemented by itself.

    Returns:    None.

    """

    current_last_id = ObjectId("000000000000000000000000")
    current_page = 0
    page_total = int(coll.find().count() / page_size)
    print("the total page : {}".format(page_total))
    data_size = 0
    while current_page <= page_total:
        print("processing the page : {}".format(current_page))
        # 查询
        condition = {"_id": {"$gt": current_last_id}}
        data = list(coll.find(condition).limit(page_size))
        # 更新 current_last_id
        current_last_id = data[-1]["_id"]
        print("current_last_id --> {}".format(current_last_id))
        # 翻页
        current_page += 1
        # 处理数据
        parse_func(data)
        data_size += len(data)

    print("the size of all processed data : --> {}".format(data_size))
    print("done.")


def aggregate_by_page_asyncio(
    coll,
    start_id: ObjectId = ObjectId("000000000000000000000000"),
    pipeline: list = [],
    session: ClientSession = None,
    options: dict = None,
    page_size: int = 100,
    parse_func: Callable[[list], None] = None,
    open_log: bool = False,
    log_file: str = "./aggregate_by_page.log",
    open_async: bool = False,
    slave_num: int = 4,
):
    """mongodb的聚合查询，具备分页查询、异步io处理数据功能。

    Args:

        coll(Collection): 目标Collection，要查询的Collection对象。
        start_id(ObjectId): 起始ObjectId。
        pipeline(list): 管道命令list。
        session(ClientSession): ClientSession对象。
        options(dict): aggregate的options选项设置。eg: {"allowDiskUse": True}
        page_size(int): 每页的数据条目数。
        parse_func(Callable): 每一个page的数据的处理函数。需要自己实现。
        open_log(bool): 是否开启日志，记录当前的Current ObjectId，方便打断任务后，再次运行时扔给start_id。
        log_file(str): 日志文件完整路径，open_log为True时需要填写，False可不填写。
        open_async(bool): 是否开启异步io处理数据，默认不开启。
        slave_num(int): 执行任务的协程数目，默认为4。

    Returns:
        None
    """

    def wprint(log_msg):
        print(log_msg)
        if open_log:
            log_file.write(log_msg + "\n")

    @asyncio.coroutine
    def parse_(chunk):
        parse_func(chunk)

    if open_log:
        log_file = open(log_file, "a+")

    current_last_id = start_id
    current_page = 0
    count = coll.find({"_id": {"$gt": current_last_id}}).count()
    page_total = (
        int(count / page_size) if count % page_size == 0 else int(count / page_size) + 1
    )
    log_msg = "# the total page : {}".format(page_total)
    print(log_msg)

    data_size = 0
    total_time = time.time()
    while current_page < page_total:
        start_time = time.time()
        log_msg = "# processing the page : {}".format(current_page)
        wprint(log_msg)
        # 查询
        condition = [
            {"$sort": {"_id": 1}},
            {"$match": {"_id": {"$gt": current_last_id}}},
            {"$limit": page_size},
        ]
        condition.extend(pipeline)
        cursor = (
            coll.aggregate(pipeline=condition, session=session, **options)
            if options
            else coll.aggregate(pipeline=condition, session=session)
        )
        data = [x for x in cursor]
        # 更新 current_last_id
        current_last_id = data[-1]["_id"]
        log_msg = "# current_last_id --> {}".format(current_last_id)
        wprint(log_msg)
        # 翻页
        current_page += 1
        # 处理数据
        if open_async:
            loop = asyncio.get_event_loop()
            chunks = gb.chunks(data, slave_num)
            tasks = list()
            for chunk in chunks:
                tasks.append(asyncio.ensure_future(parse_(chunk)))
            loop.run_until_complete(asyncio.wait(tasks))
        else:
            parse_func(data)
        data_size += len(data)
        log_msg = "# elapsed time: {}s".format(time.time() - start_time)
        wprint(log_msg)
        log_msg = "*" * 36
        wprint(log_msg)

    log_msg = "# the size of all processed data : --> {}\n".format(data_size)
    log_msg += "# total time cost is : --> {}\n".format(time.time() - total_time)
    log_msg += "# done."
    wprint(log_msg)
    if open_log:
        log_file.close()
    pass


def aggregate_by_page(
    coll,
    start_id: ObjectId = ObjectId("000000000000000000000000"),
    pipeline: list = [],
    session: ClientSession = None,
    options: dict = None,
    page_size: int = 100,
    parse_func: Callable[[list], None] = None,
    open_log: bool = False,
    log_file: str = "./aggregate_by_page.log",
):
    """mongodb的聚合查询，具备分页查询功能。

    Args:

        coll(Collection): 目标Collection，要查询的Collection对象。
        start_id(ObjectId): 起始ObjectId。
        pipeline(list): 管道命令list。
        session(ClientSession): ClientSession对象。
        options(dict): aggregate的options选项设置。eg: {"allowDiskUse": True}
        page_size(int): 每页的数据条目数。
        parse_func(Callable): 每一个page的数据的处理函数。需要自己实现。
        open_log(bool): 是否开启日志，记录当前的Current ObjectId，方便打断任务后，再次运行时扔给start_id。
        log_file(str): 日志文件完整路径，open_log为True时需要填写，False可不填写。

    Returns:
        None
    """

    def wprint(log_msg):
        print(log_msg)
        if open_log:
            log_file.write(log_msg + "\n")

    if open_log:
        log_file = open("aggregate_by_page.log", "a+")

    current_last_id = start_id
    current_page = 0
    count = coll.find({"_id": {"$gt": current_last_id}}).count()
    page_total = (
        int(count / page_size) if count % page_size == 0 else int(count / page_size) + 1
    )
    log_msg = "# the total page : {}".format(page_total)
    print(log_msg)

    data_size = 0
    total_time = time.time()

    while current_page < page_total:
        start_time = time.time()
        log_msg = "# processing the page : {}".format(current_page)
        wprint(log_msg)
        # 查询
        condition = [
            {"$sort": {"_id": 1}},
            {"$match": {"_id": {"$gt": current_last_id}}},
            {"$limit": page_size},
        ]
        condition.extend(pipeline)
        cursor = (
            coll.aggregate(pipeline=condition, session=session, **options)
            if options
            else coll.aggregate(pipeline=condition, session=session)
        )
        data = [x for x in cursor]
        # 更新 current_last_id
        current_last_id = data[-1]["_id"]
        log_msg = "# current_last_id --> {}".format(current_last_id)
        wprint(log_msg)
        # 翻页
        current_page += 1
        # 处理数据
        parse_func(data)
        data_size += len(data)
        log_msg = "# elapsed time: {}s".format(time.time() - start_time)
        wprint(log_msg)
        log_msg = "*" * 36
        wprint(log_msg)

    log_msg = "# the size of all processed data : --> {}\n".format(data_size)
    log_msg += "# total time cost is : --> {}\n".format(time.time() - total_time)
    log_msg += "# done."
    wprint(log_msg)
    if open_log:
        log_file.close()
    pass
